@@ -621,6 +621,12 @@ int main(int argc, char** argv)
 		exit(1);
 	}
 
+    gsm_util_export_activation_environment (NULL);
+
+#ifdef HAVE_SYSTEMD
+    gsm_util_export_user_environment (NULL);
+#endif
+
 	ukdm_log_init();
 
 	/* Allows to enable/disable debug from GSettings only if it is not set from argument */
